#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (C) 2015 Daniel Standage <daniel.standage@gmail.com>
#
# This file is part of tag (http://github.com/standage/tag) and is licensed
# under the BSD 3-clause license: see LICENSE.
# -----------------------------------------------------------------------------

from __future__ import print_function
import re


class Sequence():
    """
    Represents a sequence (almost always a nucleotide sequence).

    We do not encourage putting FASTA sequences in your GFF3 files, but since
    the specification explicitly allows it we have to handle it. :-(

    >>> s = Sequence('>contig1 description', 'ACGT')
    >>> s.defline
    '>contig1 description'
    >>> s.seq
    'ACGT'
    >>> s.seqid
    'contig1'
    >>> len(s)
    4
    >>> s = Sequence('>gi|12345|gb|BOGUSSEQ', 'GATTACA')
    >>> s.seq
    'GATTACA'
    >>> s.accession
    'BOGUSSEQ'
    """

    def __init__(self, defline, seq):
        assert defline.startswith('>') and defline[1] != ' '
        self.defline = defline
        self.seq = seq

    def __str__(self):
        return self.defline + '\n' + self.format_seq()

    def __repr__(self):
        return str(self)

    def __len__(self):
        return len(self.seq)

    def __lt__(self, other):
        if isinstance(other, Sequence):
            return self.seqid < other.seqid
        return False

    def __le__(self, other):
        if isinstance(other, Sequence):
            return self.seqid <= other.seqid
        return False

    def __gt__(self, other):
        if isinstance(other, Sequence):
            return self.seqid > other.seqid
        return True

    def __ge__(self, other):
        if isinstance(other, Sequence):
            return self.seqid >= other.seqid
        return True

    @property
    def seqid(self):
        return self.defline[1:].split(' ')[0]

    @property
    def accession(self):
        """
        Parse accession number from commonly supported formats.

        If the defline does not match one of the following formats, the entire
        description (sans leading caret) will be returned.

        * >gi|572257426|ref|XP_006607122.1|
        * >gnl|Tcas|XP_008191512.1
        * >lcl|PdomMRNAr1.2-10981.1
        """
        accession = None
        if self.defline.startswith('>gi|'):
            match = re.match('>gi\|\d+\|[^\|]+\|([^\|\n ]+)', self.defline)
            if match:
                accession = match.group(1)
        elif self.defline.startswith('>gnl|'):
            match = re.match('>gnl\|[^\|]+\|([^\|\n ]+)', self.defline)
            if match:
                accession = match.group(1)
        elif self.defline.startswith('>lcl|'):
            match = re.match('>lcl\|([^\|\n ]+)', self.defline)
            if match:
                accession = match.group(1)
        return accession

    def format_seq(self, outstream=None, linewidth=70):
        """
        Print a sequence in a readable format.

        :param outstream: if `None`, formatted sequence is returned as a
                          string; otherwise, it is treated as a file-like
                          object and the formatted sequence is printed to the
                          outstream
        :param linewidth: width for wrapping sequences over multiple lines; set
                          to 0 for no wrapping
        """
        if linewidth == 0 or len(self.seq) <= linewidth:
            if outstream is None:
                return self.seq
            else:
                print(self.seq, file=outstream)
                return

        i = 0
        seq = ''
        while i < len(self.seq):
            if outstream is None:
                seq += self.seq[i:i+linewidth] + '\n'
            else:
                print(self.seq[i:i+linewidth], file=outstream)
            i += linewidth
        if outstream is None:
            return seq



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (C) 2015 Daniel Standage <daniel.standage@gmail.com>
#
# This file is part of tag (http://github.com/standage/tag) and is licensed
# under the BSD 3-clause license: see LICENSE.
# -----------------------------------------------------------------------------


class Range(object):
    """
    Represents a region of a biological sequence.

    Start and end coordinates correspond to the 0-based half-open interval:
    start inclusive, end exclusive (start=0 and end=10 corresonds to the first
    10 nucleotides).

    >>> rng = Range(0, 1000)
    >>> rng
    [0, 1000)
    >>> len(rng)
    1000
    >>> rng.start = 500
    >>> rng.start
    500
    >>> rng.end
    1000
    """

    def __init__(self, start, end):
        assert start >= 0, ('start coordinate {} invalid, must be an '
                            'integer >= 0'.format(start))
        assert end >= 0, ('end coordinate {} invalid,  must be an '
                          'integer >= 0'.format(end))
        assert start <= end, ('coordinates [{}, {}] invalid, start must be '
                              '<= end'.format(start, end))
        self._start = start
        self._end = end

    def __str__(self):
        if self._start == self._end:
            return str(self._start)
        return '[{}, {})'.format(self._start, self._end)

    def __repr__(self):
        return str(self)

    def __len__(self):
        return self._end - self._start

    # Begin rich comparison operators for Python 3 support
    def __eq__(self, other):
        """
        Rich comparison operator for Python 3 support.

        Any object that defines a custom `__eq__()` operator without also
        defining a custom `__hash__()` operator is unhashable. At least for
        now, I don't see this being a problem.
        """
        return self._start == other.start and self._end == other.end

    def __lt__(self, other):
        return self._start < other.start or (self._start == other.start and
                                             self._end < other.end)

    def __le__(self, other):
        return self == other or self < other

    def __gt__(self, other):
        return self._start > other.start or (self._start == other.start and
                                             self._end > other.end)

    def __ge__(self, other):
        return self == other or self > other
    # End rich comparison operators for Python 3 support

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, newstart):
        assert newstart >= 0, \
            ('new start coordinate {} invalid, must be an '
             'integer >= 0'.format(newstart))
        assert newstart <= self._end, \
            ('new start coordinate {} invalid, must '
             'be <= end {}'.format(newstart, self._end))
        self._start = newstart

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, newend):
        assert newend >= 0, \
            ('new end coordinate {} invalid, must be an '
             'integer >= 0'.format(newend))
        assert self._start <= newend, \
            ('new end coordinate {} is invalid, '
             'must be >= start {}'.format(newend, self._start))
        self._end = newend

    def merge(self, other):
        """
        Merge this range object with another (ranges need not overlap or abut).

        :returns: a new Range object representing the interval containing both
                  ranges.
        """
        newstart = min(self._start, other.start)
        newend = max(self._end, other.end)
        return Range(newstart, newend)

    def intersect(self, other):
        """
        Determine the interval of overlap between this range and another.

        :returns: a new Range object representing the overlapping interval,
                  or `None` if the ranges do not overlap.
        """
        if not self.overlap(other):
            return None

        newstart = max(self._start, other.start)
        newend = min(self._end, other.end)
        return Range(newstart, newend)

    def overlap(self, other):
        """Determine whether this range overlaps with another."""
        if self._start < other.end and self._end > other.start:
            return True
        return False

    def contains(self, other):
        """Determine whether this range contains another."""
        if self._start <= other.start and self._end >= other.end:
            return True
        return False

    def within(self, other):
        """Determine whether this range is contained within another."""
        return other.contains(self)

    def transform(self, offset):
        """
        Shift this range by the specified offset.

        Note: the resulting range must be a valid interval.
        """
        assert self._start + offset > 0, \
            ('offset {} invalid; resulting range [{}, {}) is '
             'undefined'.format(offset, self._start+offset, self._end+offset))
        self._start += offset
        self._end += offset



#!/usr/bin/env python
#
# ------------------------------------------------------------------------------
# Copyright (C) 2015 Daniel Standage <daniel.standage@gmail.com>
#
# This file is part of tag (http://github.com/standage/tag) and is licensed
# under the BSD 3-clause license: see LICENSE.
# ------------------------------------------------------------------------------


class Comment():
    """
    Represents a comment in an annotation (GFF3) file.

    Any GFF3 entry starting with >= 1 '#' characters is treated as a comment,
    with two exceptions:

    - the separator directive, a line containing '###' and nothing more
    - any entry beginning with just two '#' characters is treated as a
      directive.
    """

    def __init__(self, data):
        assert data.startswith('#')
        self._rawdata = data

    def __repr__(self):
        return self._rawdata

    def __str__(self):
        i = 0
        while self._rawdata[i] in '# ':
            i += 1
        return self._rawdata[i:]

    def __lt__(self, other):
        from .directive import Directive
        from .feature import Feature
        if isinstance(other, Directive):
            return False
        if isinstance(other, Feature):
            return True
        return self._rawdata < other._rawdata

    def __gt__(self, other):
        from .directive import Directive
        from .feature import Feature
        if isinstance(other, Directive):
            return True
        if isinstance(other, Feature):
            return False
        return self._rawdata > other._rawdata



#!/usr/bin/env python
#
# ------------------------------------------------------------------------------
# Copyright (C) 2015 Daniel Standage <daniel.standage@gmail.com>
#
# This file is part of tag (http://github.com/standage/tag) and is licensed
# under the BSD 3-clause license: see LICENSE.
# ------------------------------------------------------------------------------

import re
from tag.range import Range


dirtypes = ['gff-version', 'sequence-region', 'feature-ontology', 'species',
            'attribute-ontology', 'source-ontology', 'genome-build']


class Directive():
    """
    Represents a directive from a GFF3 file.

    This class is primarily for error checking and data access. Once created,
    `Directive` objects should be treated as read-only: modify at your peril!
    Also, separator directives (`###`) and the `##FASTA` directive are handled
    directly by parsers and not by this class.

    Directives not explicitly declared in the GFF3 spec are application
    specific: they will be parsed without complaint, but no guarantees can be
    made about accessing their attributes.

    >>> sr = Directive('##sequence-region chr1 5000 10000')
    >>> sr.type
    'sequence-region'
    >>> sr.seqid
    'chr1'
    >>> gb = Directive('##genome-build   BeeBase 4.5')
    >>> gb.type
    'genome-build'
    >>> gb.source
    'BeeBase'
    """

    def __init__(self, data):
        assert data.startswith('##')
        self._rawdata = data

        formatmatch = re.match('##gff-version\s+(\d+)', data)
        if formatmatch:
            self.dirtype = 'gff-version'
            self.version = formatmatch.group(1)
            assert self.version == '3', 'Only GFF version 3 is supported'
            return

        formatmatch = re.match('##sequence-region\s+(\S+) (\d+) (\d+)', data)
        if formatmatch:
            self.dirtype = 'sequence-region'
            self.seqid = formatmatch.group(1)
            self.range = Range(int(formatmatch.group(2)) - 1,
                               int(formatmatch.group(3)))
            return

        formatmatch = re.match('##((feature|attribute|source)-ontology)'
                               '\s+(\S+)', data)
        if formatmatch:
            self.dirtype = formatmatch.group(1)
            self.uri = formatmatch.group(3)
            return

        formatmatch = re.match('##species\s+(\S+)', data)
        if formatmatch:
            self.dirtype = 'species'
            self.uri = formatmatch.group(1)
            return

        formatmatch = re.match('##genome-build\s+(\S+)\s+(\S+)', data)
        if formatmatch:
            self.dirtype = 'genome-build'
            self.source = formatmatch.group(1)
            self.build_name = formatmatch.group(2)
            return

        formatmatch = re.match('##(\S+)(\s+(.+))*', data)
        assert formatmatch
        self.dirtype = formatmatch.group(1)
        self.data = formatmatch.group(3)

        assert self.dirtype is not None

    @property
    def type(self):
        if self.dirtype in dirtypes:
            return self.dirtype
        return None

    def __repr__(self):
        return self._rawdata

    def __lt__(self, other):
        if self.type == 'gff-version':
            return True

        if self.type == 'sequence-region':
            if isinstance(other, Directive):
                if other.type == 'gff-version':
                    return False
                elif other.type == 'sequence-region':
                    if self.seqid == other.seqid:
                        return self.range.__lt__(other.range)
                    else:
                        return self.seqid < other.seqid
                else:
                    return True
            else:
                return True

        if isinstance(other, Directive):
            return self._rawdata < other._rawdata
        else:
            return True

    def __le__(self, other):
        if self.type == 'gff-version':
            return True

        if self.type == 'sequence-region':
            if isinstance(other, Directive):
                if other.type == 'gff-version':
                    return False
                elif other.type == 'sequence-region':
                    if self.seqid == other.seqid:
                        return self.range.__le__(other.range)
                    else:
                        return self.seqid <= other.seqid
                else:
                    return True
            else:
                return True

        if isinstance(other, Directive):
            return self._rawdata <= other._rawdata
        else:
            return True

    def __gt__(self, other):
        return not self.__le__(other)

    def __ge__(self, other):
        return not self.__lt__(other)



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (C) 2016 Daniel Standage <daniel.standage@gmail.com>
#
# This file is part of tag (http://github.com/standage/tag) and is licensed
# under the BSD 3-clause license: see LICENSE.
# -----------------------------------------------------------------------------

from __future__ import print_function
from collections import defaultdict
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
import sys
import tag
from tag.feature import Feature
from tag.reader import GFF3Reader


class GFF3Writer():
    """
    Writes sequence features and other GFF3 entries to a file.

    The :code:`instream` is expected to be an iterable of sequence features and
    other related objects. Set :code:`outfile` to :code:`-` to write output
    to stdout.

    >>> # Sort and tidy GFF3 file in 3 lines!
    >>> reader = GFF3Reader(infilename='tests/testdata/grape-cpgat.gff3')
    >>> writer = GFF3Writer(instream=reader, outfile='/dev/null')
    >>> writer.write()
    """

    def __init__(self, instream, outfile='-'):
        self._instream = instream
        self.outfilename = outfile
        self.outfile = None
        if outfile == '-':
            self.outfile == sys.stdout
        elif isinstance(outfile, StringIO):
            self.outfile = outfile
        else:
            self.outfile = tag.open(outfile, 'w')
        self.retainids = False
        self.feature_counts = defaultdict(int)

    def __del__(self):
        if self.outfilename != '-':
            self.outfile.close()

    def write(self, relax=False):
        """Pull features from the instream and write them to the output."""
        for entry in self._instream:
            if isinstance(entry, Feature):
                for feature in entry:
                    if feature.num_children > 0:
                        self.feature_counts[feature.type] += 1
                        fid = '{}{}'.format(feature.type,
                                            self.feature_counts[feature.type])
                        feature.add_attribute('ID', fid)
                    else:
                        feature.drop_attribute('ID')
            print(repr(entry), file=self.outfile)



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (C) 2015 Daniel Standage <daniel.standage@gmail.com>
#
# This file is part of tag (http://github.com/standage/tag) and is licensed
# under the BSD 3-clause license: see LICENSE.
# -----------------------------------------------------------------------------
"""Package-wide configuration"""

try:
    import __builtin__ as builtins
except:
    import builtins
from . import comment
from . import feature
from . import range
from . import reader
from . import writer
from gzip import open as gzopen

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions


def open(filename, mode):
    if mode not in ['r', 'w']:
        raise ValueError('invalid mode "{}"'.format(mode))
    openfunc = builtins.open
    if filename.endswith('.gz'):
        openfunc = gzopen
        mode += 't'
    return openfunc(filename, mode)


def demo_feature():
    gene = feature.Feature(
        'contig1\tsnap\tgene\t1000\t7500\t.\t+\t.\tID=gene1'
    )
    mrna = feature.Feature(
        'contig1\tsnap\tmRNA\t1000\t7500\t.\t+\t.\tID=mrna1;Parent=gene1'
    )
    exon1 = feature.Feature(
        'contig1\tsnap\texon\t1000\t3700\t.\t+\t.\tParent=mrna1'
    )
    exon2 = feature.Feature(
        'contig1\tsnap\texon\t7250\t7500\t.\t+\t.\tParent=mrna1'
    )
    cds1 = feature.Feature(
        'contig1\tsnap\tCDS\t1289\t3700\t.\t+\t0\tID=cds1;Parent=mrna1'
    )
    cds2 = feature.Feature(
        'contig1\tsnap\tCDS\t7250\t7352\t.\t+\t0\tID=cds1;Parent=mrna1'
    )
    cds1.add_sibling(cds2)
    for f in [exon1, exon2, cds1, cds2]:
        mrna.add_child(f)
    gene.add_child(mrna)
    return gene



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (C) 2015 Daniel Standage <daniel.standage@gmail.com>
#
# This file is part of tag (http://github.com/standage/tag) and is licensed
# under the BSD 3-clause license: see LICENSE.
# -----------------------------------------------------------------------------

from __future__ import print_function
import argparse
import intervaltree
import tag


def interval_set_span(intset):
    begin = min([x for x, y, z in intset])
    end = max([y for x, y, z in intset])
    cds = set([z for x, y, z in intset])
    return begin, end, cds

parser = argparse.ArgumentParser()
parser.add_argument('gff3', type=argparse.FileType('r'))
args = parser.parse_args()

coding_exons = dict()
reader = tag.reader.GFF3Reader(instream=args.gff3)
for record in reader:
    if not isinstance(record, tag.feature.Feature):
        continue

    for feature in record:
        if feature.type != 'mRNA':
            continue

        for subfeat in feature:
            if subfeat.type == 'CDS':
                if subfeat.seqid not in coding_exons:
                    coding_exons[subfeat.seqid] = intervaltree.IntervalTree()
                coding_exons[subfeat.seqid].addi(subfeat.start - 1,
                                                 subfeat.end,
                                                 subfeat)

total_occ = 0
ints_acct_for = dict()
for seqid in coding_exons:
    ints_acct_for[seqid] = intervaltree.IntervalTree()
    for cds_interval in coding_exons[seqid]:
        begin, end, cds = cds_interval
        if ints_acct_for[seqid][begin:end] != set():
            continue

        cds = set(cds)
        overlapping = coding_exons[seqid][begin:end]
        testbegin, testend, testcds = interval_set_span(overlapping)
        while set(cds) < testcds:
            begin, end, cds = testbegin, testend, testcds
            overlapping = coding_exons[seqid][begin:end]
            testbegin, testend, testcds = interval_set_span(overlapping)
        total_occ += end - begin
        ints_acct_for[seqid].addi(begin, end, cds)
print(total_occ)



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (C) 2015 Daniel Standage <daniel.standage@gmail.com>
#
# This file is part of tag (http://github.com/standage/tag) and is licensed
# under the BSD 3-clause license: see LICENSE.
# -----------------------------------------------------------------------------

import argparse
import tag

parser = argparse.ArgumentParser()
parser.add_argument('-o', '--out', metavar='FILE', default='/dev/stdout',
                    help='write output to the specified file; default is '
                    'terminal (stdout)')
parser.add_argument('gff3', help='input file in GFF3 format')
args = parser.parse_args()

infile = tag.open(args.gff3, 'r')
reader = tag.reader.GFF3Reader(infile)
writer = tag.writer.GFF3Writer(reader, args.out)

writer.write()



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (C) 2015 Daniel Standage <daniel.standage@gmail.com>
#
# This file is part of tag (http://github.com/standage/tag) and is licensed
# under the BSD 3-clause license: see LICENSE.
# -----------------------------------------------------------------------------

import sys
import tag
from tag.range import Range
from tag.comment import Comment
from tag.directive import Directive
from tag.feature import Feature
from tag.sequence import Sequence


def parse_fasta(data):  # pragma: no cover
    """
    Load sequences in Fasta format.

    This generator function yields a Sequence object for each sequence record
    in a GFF3 file. Implementation stolen shamelessly from
    http://stackoverflow.com/a/7655072/459780.
    """
    name, seq = None, []
    for line in data:
        line = line.rstrip()
        if line.startswith('>'):
            if name:
                yield Sequence(name, ''.join(seq))
            name, seq = line, []
        else:
            seq.append(line)
    if name:
        yield Sequence(name, ''.join(seq))


class GFF3Reader():
    """
    Loads sequence features and other GFF3 entries into memory.

    Features and other objects are obtained by iterating over the GFF3Reader
    object.

    >>> reader = GFF3Reader(infilename='tests/testdata/grape-cpgat.gff3')
    >>> for entry in reader:
    ...     if hasattr(entry, 'type') and entry.type == 'gene':
    ...         print(entry.slug)
    gene@chr8[72, 5081]
    gene@chr8[10538, 11678]
    gene@chr8[22053, 23448]

    By default, the GFF3Reader assumes the input is not in sorted order and
    loads the entire annotation into memory to ensure sorting and resolution
    of ID/Parent relationships. If the input is sorted according to genomic
    position and independent features are separated by :code:`###` directives,
    memory consumption will be drastically reduced by setting the
    :code:`assumesorted` attribute to True.

    The :code:`strict` attribute enforces some additional sanity checks, which
    in some exceptional cases may need to be relaxed.
    """

    def __init__(self, instream=None, infilename=None,
                 assumesorted=False, strict=True):
        assert (not instream) != (not infilename), ('provide either an '
                                                    'instream or an infile '
                                                    'name, not both')
        self.instream = instream
        if infilename:
            self.infilename = infilename
            self.instream = tag.open(infilename, 'r')
        self.assumesorted = assumesorted
        self.strict = strict
        self.declared_regions = dict()
        self.inferred_regions = dict()
        self._counter = 0
        self._prevrecord = None

    def __iter__(self):
        """Generator function returns GFF3 entries."""
        self._reset()

        for line in self.instream:
            line = line.rstrip()
            if line == '':
                continue
            elif line == '###':
                if self.assumesorted:
                    for obj in self._resolve_features():
                        if self._prevrecord and self._prevrecord > obj:
                            msg = 'sorting error: '
                            msg += '{} > {}'.format(self._prevrecord, obj)
                            raise ValueError(msg)
                        self._prevrecord = obj
                        self._counter += 1
                        yield obj
            elif line.startswith('#'):
                if line == '##FASTA':
                    for sequence in parse_fasta(self.instream):
                        self.records.append(sequence)
                    break
                elif line.startswith('##') and line[2] != '#':
                    record = Directive(line)
                    if record.type == 'sequence-region':
                        if record.seqid in self.declared_regions:
                            m = ('sequence {} declared in multiple ##sequence'
                                 '-region entries'.format(record.seqid))
                            raise ValueError(m)
                        self.declared_regions[record.seqid] = record
                else:
                    record = Comment(line)
                self.records.append(record)
            else:
                feature = Feature(line)

                if feature.seqid in self.declared_regions:
                    seqregion = self.declared_regions[feature.seqid]
                    if not feature._range.within(seqregion.range):
                        msg = 'feature {} out-of-bounds'.format(feature.slug)
                        raise ValueError(msg)

                if feature.seqid not in self.inferred_regions:
                    self.inferred_regions[feature.seqid] = Range(feature.start,
                                                                 feature.end)
                rstr = self.inferred_regions[feature.seqid].start
                rend = self.inferred_regions[feature.seqid].end
                fstr = feature.start
                fend = feature.end
                self.inferred_regions[feature.seqid].start = min(rstr, fstr)
                self.inferred_regions[feature.seqid].end = max(rend, fend)

                parentid = feature.get_attribute('Parent')
                if parentid is None:
                    # All components of a multi-feature are added here...
                    self.records.append(feature)
                else:
                    if parentid not in self.featsbyparent:
                        self.featsbyparent[parentid] = list()
                    self.featsbyparent[parentid].append(feature)

                featureid = feature.get_attribute('ID')
                if featureid is not None:
                    if featureid in self.featsbyid:
                        # Validate multi-features
                        other = self.featsbyid[featureid]
                        assert feature.type == other.type, \
                            'feature type disagreement for ID="%s": %s vs %s' \
                            % (featureid, feature.type, other.type)
                        other.add_sibling(feature)
                    else:
                        self.featsbyid[featureid] = feature

        for obj in self._resolve_features():
            if isinstance(obj, Feature):
                if obj.is_multi and obj.multi_rep != obj:
                    # ...so we must filter multi-features here and only yield
                    # the multi-feature representative
                    continue

            if self._counter == 0:
                isv = isinstance(obj, Directive) and obj.type == 'gff-version'
                if not isv:
                    self._prevrecord = Directive('##gff-version 3')
                    self._counter += 1
                    yield self._prevrecord

            self._prevrecord = obj
            self._counter += 1
            yield obj

    def _resolve_features(self):
        """Resolve Parent/ID relationships and yield all top-level features."""

        for parentid in self.featsbyparent:
            parent = self.featsbyid[parentid]
            for child in self.featsbyparent[parentid]:
                parent.add_child(child, rangecheck=self.strict)

        if not self.assumesorted:
            for seqid in self.inferred_regions:
                if seqid not in self.declared_regions:
                    seqrange = self.inferred_regions[seqid]
                    srstring = '##sequence-region {:s} {:d} {:d}'.format(
                        seqid, seqrange.start + 1, seqrange.end
                    )
                    seqregion = Directive(srstring)
                    self.records.append(seqregion)

        for record in sorted(self.records):
            yield record
        self._reset()

    def _reset(self):
        """Clear internal data structure."""
        self.records = list()
        self.featsbyid = dict()
        self.featsbyparent = dict()
        self.countsbytype = dict()



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (C) 2015 Daniel Standage <daniel.standage@gmail.com>
#
# This file is part of tag (http://github.com/standage/tag) and is licensed
# under the BSD 3-clause license: see LICENSE.
# -----------------------------------------------------------------------------

from setuptools import setup
import glob
import versioneer


setup(name='tag',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      description='Toolkit for genome annotation data analysis',
      url='http://github.com/standage/tag',
      author='Daniel Standage',
      author_email='daniel.standage@gmail.com',
      license='BSD-3',
      packages=['tag'],
      tests_require=['pytest', 'pytest-cov', 'pep8'],
      scripts=list(glob.glob('scripts/*.py')),
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'License :: OSI Approved :: BSD License',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Topic :: Scientific/Engineering :: Bio-Informatics'
      ],
      zip_safe=True)



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (C) 2015 Daniel Standage <daniel.standage@gmail.com>
#
# This file is part of tag (http://github.com/standage/tag) and is licensed
# under the BSD 3-clause license: see LICENSE.
# -----------------------------------------------------------------------------

import tag
from tag.comment import Comment
from tag.directive import Directive
from tag.range import Range
from tag.sequence import Sequence


class Feature(object):
    """
    Represents a feature entry from a GFF3 file.

    >>> feature = tag.demo_feature()
    >>> feature.seqid
    'contig1'
    >>> feature.source
    'snap'
    >>> feature.type
    'gene'
    >>> feature.start, feature.end
    (999, 7500)
    >>> feature.score is None
    True
    >>> feature.strand
    '+'
    >>> feature.phase is None
    True
    >>> feature.attributes
    'ID=gene1'
    >>> feature.num_children
    1
    >>> feature.is_multi
    False
    >>> feature.is_toplevel
    True
    >>> for child in feature:
    ...     if child.type == 'CDS':
    ...         assert child.get_attribute('ID') == 'cds1'
    >>> feature.slug
    'gene@contig1[1000, 7500]'
    """

    def __init__(self, data):
        fields = data.split('\t')
        assert len(fields) == 9
        self.children = None
        self.multi_rep = None
        self.siblings = None

        self._seqid = fields[0]
        self._source = fields[1]
        self._type = fields[2]
        self._range = Range(int(fields[3]) - 1, int(fields[4]))
        if fields[5] == '.':
            self.score = None
        else:
            self.score = float(fields[5])
        self._strand = fields[6]
        if fields[7] == '.':
            self.phase = None
        else:
            self.phase = int(fields[7])
        self._attrs = self.parse_attributes(fields[8])

        assert self._strand in ['+', '-', '.'], \
            'invalid strand "{}"'.format(self._strand)
        if self.phase is not None:
            assert self.phase in [0, 1, 2], \
                'invalid phase "{}"'.format(self.phase)

    def __str__(self):
        """String representation of the feature, sans children."""
        score = '.'
        if self.score is not None:
            score = "{:.3f}".format(self.score)
        phase = '.'
        if self.phase is not None:
            phase = str(self.phase)
        return '\t'.join([
            self.seqid, self.source, self.type, str(self.start + 1),
            str(self.end), score, self.strand, phase, self.attributes
        ])

    def __repr__(self):
        """Full representation of the feature, with children"""
        string = ''
        for feature in self:
            if string != '':
                string += '\n'
            string += str(feature)
        if self.children is not None or self.is_multi:
            string += '\n###'
        return string

    def __len__(self):
        return len(self._range)

    def __lt__(self, other):
        if isinstance(other, Directive) or isinstance(other, Comment):
            return False
        elif isinstance(other, Sequence):
            return True
        assert isinstance(other, Feature)

        if self.seqid < other.seqid:
            return True
        elif self.seqid > other.seqid:
            return False
        elif self._range == other._range:
            return self.type > other.type
        return self._range < other._range

    def __le__(self, other):
        if isinstance(other, Directive) or isinstance(other, Comment):
            return False
        elif isinstance(other, Sequence):
            return True
        assert isinstance(other, Feature)

        if self.seqid < other.seqid:
            return True
        elif self.seqid > other.seqid:
            return False
        elif self._range == other._range:
            return self.type >= other.type
        return self._range <= other._range

    def __gt__(self, other):
        if isinstance(other, Directive) or isinstance(other, Comment):
            return True
        elif isinstance(other, Sequence):
            return False
        assert isinstance(other, Feature)

        if self.seqid > other.seqid:
            return True
        elif self.seqid < other.seqid:
            return False
        elif self._range == other._range:
            return self.type < other.type
        return self._range > other._range

    def __ge__(self, other):
        if isinstance(other, Directive) or isinstance(other, Comment):
            return True
        elif isinstance(other, Sequence):
            return False
        assert isinstance(other, Feature)

        if self.seqid > other.seqid:
            return True
        elif self.seqid < other.seqid:
            return False
        elif self._range == other._range:
            return self.type <= other.type
        return self._range >= other._range

    def __iter__(self):
        """Generator iterates through a feature and all its subfeatures."""
        sorted_features = list()
        self._visit(L=sorted_features, marked={}, tempmarked={})
        for feat in sorted_features:
            yield feat

    def _visit(self, L, marked, tempmarked):
        """
        Sort features topologically.

        This recursive function uses depth-first search to find an ordering of
        the features in the feature graph that is sorted both topologically and
        with respect to genome coordinates.

        Implementation based on Wikipedia's description of the algorithm in
        Cormen's *Introduction to Algorithms*.
        http://en.wikipedia.org/wiki/Topological_sorting#Algorithms

        There are potentially many valid topological sorts of a feature graph,
        but only one that is also sorted with respect to genome coordinates
        (excluding different orderings of, for example, exons and CDS features
        with the same coordinates). Iterating through feature children in
        reversed order (in this functions' inner-most loop) seems to be the key
        to sorting with respect to genome coordinates.
        """
        if self in tempmarked:
            raise Exception('feature graph is cyclic')
        if self not in marked:
            tempmarked[self] = True
            features = list()
            if self.siblings is not None and self.is_toplevel:
                features.extend(reversed(self.siblings))
            if self.children is not None:
                features.extend(reversed(self.children))
            if len(features) > 0:
                for feature in features:
                    feature._visit(L, marked, tempmarked)
            marked[self] = True
            del tempmarked[self]
            L.insert(0, self)

    def add_child(self, child, rangecheck=False):
        """Add a child feature to this feature."""
        assert self.seqid == child.seqid, \
            (
                'seqid mismatch for feature {} ({} vs {})'.format(
                    self.fid, self.seqid, child.seqid
                )
            )
        if rangecheck is True:
            assert self._strand == child._strand, \
                ('child of feature {} has a different strand'.format(self.fid))
            assert self._range.contains(child._range), \
                (
                    'child of feature {} is not contained within its span '
                    '({}-{})'.format(self.fid, child.start, child.end)
                )
        if self.children is None:
            self.children = list()
        self.children.append(child)
        self.children.sort()

    @property
    def num_children(self):
        if self.children is None:
            return 0
        return len(self.children)

    @property
    def fid(self):
        return self.get_attribute('ID')

    @property
    def slug(self):
        """
        A concise slug for this feature.

        Unlike the internal representation, which is 0-based half-open, the
        slug is a 1-based closed interval (a la GFF3).
        """
        return '{:s}@{:s}[{:d}, {:d}]'.format(self.type, self.seqid,
                                              self.start + 1, self.end)

    @property
    def is_multi(self):
        return self.multi_rep is not None

    @property
    def is_toplevel(self):
        return self.get_attribute('Parent') is None

    def add_sibling(self, sibling):
        """
        Designate this a multi-feature representative and add a co-feature.

        Some features exist discontinuously on the sequence, and therefore
        cannot be declared with a single GFF3 entry (which can encode only a
        single interval). The canonical encoding for these types of features is
        called a multi-feature, in which a single feature is declared on
        multiple lines with multiple entries all sharing the same feature type
        and ID attribute. This is commonly done with coding sequence (CDS)
        features.

        In this package, each multi-feature has a single "representative"
        feature object, and all other objects/entries associated with that
        multi-feature are attached to it as "siblings".

        Invoking this method will designate the calling feature as the
        multi-feature representative and add the argument as a sibling.
        """
        if self.siblings is None:
            self.siblings = list()
            self.multi_rep = self
        sibling.multi_rep = self
        self.siblings.append(sibling)

    @property
    def seqid(self):
        return self._seqid

    @seqid.setter
    def seqid(self, newseqid):
        """When modifying seqid, make sure to update seqid of children also."""
        for feature in self:
            feature._seqid = newseqid

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, newsource):
        """When modifying source, also update children with matching source."""
        oldsource = self.source
        for feature in self:
            if feature.source == oldsource:
                feature._source = newsource

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, newtype):
        """If the feature is a multifeature, update all entries."""
        self._type = newtype
        if self.is_multi:
            for sibling in self.multi_rep.siblings:
                sibling._type = newtype

    @property
    def start(self):
        return self._range.start

    @property
    def end(self):
        return self._range.end

    def set_coord(self, start, end):
        """Manually reset the feature's coordinates."""
        self._range = Range(start, end)

    def transform(self, offset, newseqid=None):
        """Transform the feature's coordinates by the given offset."""
        for feature in self:
            feature._range.transform(offset)
            if newseqid is not None:
                feature.seqid = newseqid

    @property
    def strand(self):
        return self._strand

    @property
    def attributes(self):
        if len(self._attrs) == 0:
            return '.'

        attrs = list()
        if 'ID' in self._attrs:
            attrs.append('ID=' + self.get_attribute('ID'))
        if 'Parent' in self._attrs:
            parent = self.get_attribute('Parent', as_string=True)
            attrs.append('Parent=' + parent)
        if 'Name' in self._attrs:
            name = self.get_attribute('Name', as_string=True)
            attrs.append('Name=' + name)
        for attrkey in sorted(self._attrs):
            if attrkey in ['ID', 'Parent', 'Name']:
                continue
            value = self.get_attribute(attrkey, as_string=True)
            attrs.append('{}={}'.format(attrkey, value))
        return ';'.join(attrs)

    def add_attribute(self, attrkey, attrvalue, append=False, oldvalue=None):
        """
        Add an attribute to this feature.

        Feature attributes are stored as nested dictionaries.

        Each feature can only have one ID, so ID attribute mapping is 'string'
        to 'string'. All other attributes can have multiple values, so mapping
        is 'string' to 'dict of strings'.

        By default, adding an attribute that already exists will cause the old
        value to be overwritten. If the `append` option is true, the new
        attribute value will not overwrite the old value, but will be appended
        as a second value. (Note: ID attributes can have only 1 value.)

        If the `oldvalue` option is set, the new value will replace the old
        value. This is necessary for updating an attribute that has multiple
        values without completely overwriting all old values. (Note: The
        `append` option is ignored when `oldvalue` is set.)
        """
        # Handle ID/Parent relationships
        if attrkey == 'ID':
            if self.children is not None:
                oldid = self.get_attribute('ID')
                for child in self.children:
                    child.add_attribute('Parent', attrvalue,
                                        oldvalue=oldid)
            self._attrs[attrkey] = attrvalue
            return

        # Handle all other attribute types
        if oldvalue is not None:
            if attrkey in self._attrs:
                assert oldvalue in self._attrs[attrkey]
                del self._attrs[attrkey][oldvalue]
        if attrkey not in self._attrs or append is False:
            self._attrs[attrkey] = dict()
        self._attrs[attrkey][attrvalue] = True

    def get_attribute(self, attrkey, as_string=False, as_list=False):
        """
        Get the value of an attribute.

        By default, returns a string for ID and attributes with a single value,
        and a list of strings for attributes with multiple values. The
        `as_string` and `as_list` options can be used to force the function to
        return values as a string (comma-separated in case of multiple values)
        or a list.
        """
        assert not as_string or not as_list
        if attrkey not in self._attrs:
            return None
        if attrkey == 'ID':
            return self._attrs[attrkey]
        attrvalues = list(self._attrs[attrkey])
        attrvalues.sort()
        if len(attrvalues) == 1 and not as_list:
            return attrvalues[0]
        elif as_string:
            return ','.join(attrvalues)
        return attrvalues

    def drop_attribute(self, attrkey):
        """Drop the specified attribute from the feature."""
        if attrkey in self._attrs:
            del self._attrs[attrkey]

    def get_attribute_keys(self):
        """Return a list of all this feature's attribute keys."""
        return sorted(list(self._attrs))

    def parse_attributes(self, attrstring):
        """
        Parse an attribute string.

        Given a string with semicolon-separated key-value pairs, populate a
        dictionary with the given attributes.
        """
        if attrstring == '.':
            return dict()

        attributes = dict()
        keyvaluepairs = attrstring.split(';')
        for kvp in keyvaluepairs:
            if kvp == '':
                continue
            key, value = kvp.split('=')
            if key == 'ID':
                assert ',' not in value
                attributes[key] = value
                continue
            values = value.split(',')
            valdict = dict((val, True) for val in values)
            attributes[key] = valdict
        return attributes
